use crate::felinet::handler_client::HandlerClient;

use config::Config;
use felinet::PingRequest;
use gate::{beacon_forever, felinet_send_forever};
use std::{str::FromStr, time::Duration};
use tokio::sync::mpsc;
use tonic::{transport::Endpoint, Request};
use uuid::Uuid;

mod config;
mod decoder;
mod gate;
mod util;

const MAX_PACKET_LEN: usize = 8191;

pub mod felinet {
    tonic::include_proto!("felinet");
}

#[tokio::main]
async fn main() -> anyhow::Result<()> {
    let config = Config::load()?;

    gate_forever(config).await?;

    Ok(())
}

async fn gate_forever(config: Config) -> anyhow::Result<()> {
    let uuid = Uuid::new_v4();

    let (tx, rx) = mpsc::channel(16);

    if let Some(felinet_config) = &config.felinet {
        let endpoint = Endpoint::from_str(&felinet_config.address)?
            .keep_alive_while_idle(true)
            .keep_alive_timeout(Duration::from_secs(5))
            .http2_keep_alive_interval(Duration::from_secs(5))
            .tcp_keepalive(Some(Duration::from_secs(5)));

        let client = HandlerClient::connect(endpoint).await?;

        {
            let client = client.clone();
            tokio::task::spawn(async {
                felinet_send_forever(client, rx).await;
            });
        }

        beacon_forever(felinet_config, tx.clone(), uuid)?;

        {
            let mut client = client.clone();
            tokio::task::spawn(async move {
                loop {
                    client
                        .ping(Request::new(PingRequest { uuid: uuid.into() }))
                        .await
                        .ok();

                    tokio::time::sleep(Duration::from_secs(600)).await;
                }
            });
        }
    } else {
        eprintln!(
            "No FELINET section specified in configuration. Received packets will not be uploaded!"
        );
    }

    tokio::task::spawn_blocking(move || {
        decoder::decode_forever(tx, &config.felinet, &uuid);
    })
    .await?;

    Ok(())
}
